# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for performing parsing of web pages."""
from __future__ import annotations

import asyncio
import collections
import dataclasses

import aiohttp
import bs4

from googleads_housekeeper.domain.external_parsers import base_parser


@dataclasses.dataclass
class WebsiteInfo(base_parser.EntityInfo):
    """Contains meta information of websites.

    Attributes:
        placement: URL of website's domain.
        title: Title of website.
        description: Website description from 'meta' tags.
        keywords: Website keywords from 'meta' tags.
        is_processed: Whether website was successfully parsed.
    """
    placement: str
    title: str = ''
    description: str = ''
    keywords: str = ''
    is_processed: bool = False


class WebSiteParser(base_parser.BaseParser):
    """Performs website parsing."""

    def parse(self, placements: collections.abc.Sequence[str]) -> list[WebsiteInfo]:
        """Parses provided websites.

        Args:
            placements: Sequence of website domains.
        Returns:
            Parsed information in WebsiteInfo format.
        """
        return asyncio.run(self._parse_placements(placements))

    async def _parse_placements(self,
                                placements: collections.abc.Sequence[str]) -> list[WebsiteInfo]:
        """Parses provided websites asynchronously.

        Args:
            placements: Sequence of website domains.
        Returns:
            Parsed information in WebsiteInfo format.
        """
        timeout = aiohttp.ClientTimeout(total=10)
        async with aiohttp.ClientSession(timeout=timeout) as session:
            results = []
            for placement in placements:
                results.append(self._parse_placement(placement, session))
            values = await asyncio.gather(*results)
        return values

    async def _parse_placement(self, placement: str,
                               session: aiohttp.ClientSession) -> WebsiteInfo:
        """Parses single website.

        Args:
            placement: Website domain.
        Returns:
            Parsed information in WebsiteInfo format.
        """
        url = self._convert_placement_to_url(placement)
        try:
            html = await self._fetch_html(url, session)
            placement_info = await self._extract_from_html(placement, html)
            return placement_info
        except Exception:
            return WebsiteInfo(placement=placement)

    def _convert_placement_to_url(self, placement: str) -> str:
        """Normalizes url."""
        if 'http' not in placement:
            url = f'https://{placement}'
        else:
            url = placement
        return url

    async def _fetch_html(self, url: str, session: aiohttp.ClientSession):
        """Gets website content."""
        response = await session.request(method='GET', url=url)
        response.raise_for_status()
        html = await response.text()
        return html

    async def _extract_from_html(self, url: str, html: str) -> WebsiteInfo:
        """Extracts tags from html."""
        soup = bs4.BeautifulSoup(html, 'html.parser')
        keywords = soup.find('meta', attrs={'name': 'keywords'})
        description = soup.find('meta', attrs={'name': 'description'})
        return WebsiteInfo(
            placement=url,
            title=soup.title.string if soup.title else None,
            keywords=keywords.get('content') if keywords else None,
            description=description.get('content') if description else None,
            is_processed=True)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from dataclasses import dataclass


class Event:
    ...


@dataclass
class TaskCreated(Event):
    task_id: str


@dataclass
class TaskWithScheduleCreated(Event):
    task_id: str
    task_name: str
    schedule: str


@dataclass
class TaskUpdated(Event):
    task_id: str


@dataclass
class TaskScheduleUpdated(Event):
    task_id: str
    schedule: str


@dataclass
class TaskRun(Event):
    task_id: str


@dataclass
class TaskDeleted(Event):
    task_id: str


@dataclass
class TaskScheduleDeleted(Event):
    task_id: str


@dataclass
class EntityExcluded(Event):
    criterion_id: int
    criterion_type: str
    customer_id: int
    level_id: int  # can be ad_group, campaign_id, or customer_id


@dataclass
class EntityAllowlisted(Event):
    criterion_id: int
    criterion_type: str
    customer_id: int
    level_id: int  # can be ad_group, campaign_id, or customer_id



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for building and applying various specification.

ExclusionSpecification holds an expression is checked against a particular
GaarfRow object to verify whether or not this expression is true.
"""
from __future__ import annotations

import dataclasses
import itertools
import logging
import re
from collections.abc import Sequence

from gaarf.report import GaarfReport
from gaarf.report import GaarfRow

from googleads_housekeeper.domain.core import rules_parser
from googleads_housekeeper.services import enums

_PARSEABLE_PLACEMENT_TYPES = ('WEBSITE', 'YOUTUBE_VIDEO', 'YOUTUBE_CHANNEL')
_PARSEABLE_PLACEMENT_TYPES_ENUMS = {
    enums.ExclusionTypeEnum.WEBSITE_INFO,
    enums.ExclusionTypeEnum.YOUTUBE_VIDEO_INFO,
    enums.ExclusionTypeEnum.YOUTUBE_CHANNEL_INFO
}


@dataclasses.dataclass(frozen=True)
class RuntimeOptions:
    """Defines options for applying specifications.

    Attributes:
        is_regular_query: Whether Gaarf query does not contain conversion split.
        is_conversion_query: Whether Gaarf query contains conversion split.
        conversion_name: Name(s) of conversion(s) for conversion query.
        conversion_rules: All rules related to conversion split.
    """
    is_regular_query: bool = False
    is_conversion_query: bool = False
    conversion_name: str = ''
    conversion_rules: list[BaseExclusionSpecificationEntry] = dataclasses.field(
        default_factory=list)


class BaseExclusionSpecificationEntry:
    """Base class for holding logic applicable to all specifications.

    Attributes:
        name:
            Name of the metric/dimension from Google Ads.
        operator:
            Comparator used to evaluate the expression.
        value:
            Expected value metric should take.
        exclusion_type:
            Type of the parseable entity, depending on it additional data might
            be fetched from the repository.
    """

    def __init__(self,
                 expression: str,
                 exclusion_type: enums.ExclusionTypeEnum) -> None:
        """Constructor for the class.

        Args:
            expression: Exclusion expression in a form of `name > value`.
            exclusion_type: Type of the exclusion.
        """
        elements = [
            element.strip() for element in expression.split(' ', maxsplit=2)
        ]
        if len(elements) != 3:
            raise ValueError(
                "Incorrect expression, specify in 'name > value' format")
        if elements[1] not in ('>', '>=', '<', '<=', '=', '!=', 'regexp',
                               'contains', 'has_latin_letters'):
            raise ValueError(
                'Incorrect operator for expression, '
                "only '>', '>=', '<', '<=', '=', '!=', 'regexp', 'contains' ",
                "'has_latin_letters' are supported'")

        self.name = elements[0]
        self.operator = '==' if elements[1] == '=' else elements[1]
        self.__raw_value = elements[2].replace("'", '').replace('"', '')
        if self.__raw_value.lower() == 'true':
            self.value = '^[a-zA-Z ]*$'
        elif self.__raw_value.lower() == 'false':
            self.value = '^[^a-zA-Z]*$'
        else:
            self.value = self.__raw_value
        self.exclusion_type = exclusion_type

    @property
    def corresponding_placement_type(self) -> enums.PlacementTypeEnum | None:
        if self.exclusion_type == enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
            return None
        return (enums.PlacementTypeEnum[self.exclusion_type.name.replace(
            '_INFO', '')])

    @property
    def repository_name(self) -> str | None:
        if self.exclusion_type == enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
            return None
        return self.exclusion_type.name.lower()

    def is_satisfied_by(self, placement_info: GaarfRow) -> tuple[bool, dict]:
        """Verify whether given entity satisfies stored expression.

        Args:
            placement_info: GaarfRow object that contains entity data.
        Returns:
            Tuple with results of evaluation and all necessary information on
            placement (formatted as a dict).
        """
        placement_as_dict = {}
        if (placement_info.placement_type in _PARSEABLE_PLACEMENT_TYPES and
                self.exclusion_type != enums.ExclusionTypeEnum.GOOGLE_ADS_INFO):
            placement = placement_info.extra_info.get(
                self.exclusion_type.name.lower(), {})
            if not placement:
                return False, {}
            placement_as_dict = dataclasses.asdict(placement)
        else:
            placement = placement_info
        if not hasattr(placement, self.name):
            raise ValueError(f'{self.name} is not found!')
        if hasattr(placement, 'is_processed') and not placement.is_processed:
            logging.debug(
                'Cannot get internal information on %s placement of type %s',
                placement_info.placement, placement_info.placement_type)
            return False, placement_as_dict
        if self.operator in ('regexp', 'contains', 'has_latin_letters'):
            return self._check_regexp(placement), placement_as_dict
        return self._eval_expression(placement), placement_as_dict

    def _check_regexp(self, placement: GaarfRow) -> bool:
        if placement_element := getattr(placement, self.name):
            punctuation_regex = re.compile('f[{re.escape(punctuation)}]')
            return bool(
                re.search(fr'{self.value}',
                          punctuation_regex.sub('', placement_element),
                          re.IGNORECASE))
        return False

    def _eval_expression(self, placement):
        try:
            value = float(self.value)
        except ValueError:
            value = self.value
        if isinstance(value, float):
            return eval(
                f'{getattr(placement, self.name)}{self.operator} {value}')
        return getattr(placement, self.name) == value

    def __str__(self):
        return f'{self.exclusion_type.name}:{self.name} {self.operator} {self.value}'

    def __repr__(self):
        return (f'{self.__class__.__name__}'
                f"(exclusion_type='{self.exclusion_type.name}', "
                f"name='{self.name}', "
                f"operator='{self.operator}', value='{self.value}')")

    def __eq__(self, other):
        return (self.exclusion_type, self.name, self.operator,
                self.value) == (other.exclusion_type, other.name,
                                other.operator, other.value)


class AdsExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
    """Stores Google Ads specific rules."""

    def __init__(self, expression):
        super().__init__(
            expression, exclusion_type=enums.ExclusionTypeEnum.GOOGLE_ADS_INFO)


class ContentExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
    """Stores Website specific rules."""

    def __init__(self, expression) -> None:
        super().__init__(
            expression=expression,
            exclusion_type=enums.ExclusionTypeEnum.WEBSITE_INFO)


class YouTubeChannelExclusionSpecificationEntry(BaseExclusionSpecificationEntry
                                               ):
    """Stores YouTube Channel specific rules."""

    def __init__(self, expression) -> None:
        super().__init__(
            expression=expression,
            exclusion_type=enums.ExclusionTypeEnum.YOUTUBE_CHANNEL_INFO)


class YouTubeVideoExclusionSpecificationEntry(BaseExclusionSpecificationEntry):
    """Stores YouTube Video specific rules."""

    def __init__(self, expression) -> None:
        super().__init__(
            expression=expression,
            exclusion_type=enums.ExclusionTypeEnum.YOUTUBE_VIDEO_INFO)


def create_exclusion_specification_entry(
        specification_type: str,
        condition: str) -> BaseExclusionSpecificationEntry:
    """Builds concrete specification entry class based on type.

    Args:
        specification_type: Type of desired specification entry.
        condition: Expression to use for building the specification entry.

    Returns:
        Any subclass of instance of BaseExclusionSpecificationEntry.
    """
    if specification_type == 'GOOGLE_ADS_INFO':
        return AdsExclusionSpecificationEntry(condition)
    if specification_type == 'WEBSITE_INFO':
        return ContentExclusionSpecificationEntry(condition)
    if specification_type == 'YOUTUBE_CHANNEL_INFO':
        return YouTubeChannelExclusionSpecificationEntry(condition)
    if specification_type == 'YOUTUBE_VIDEO_INFO':
        return YouTubeVideoExclusionSpecificationEntry(condition)
    raise ValueError(f'Incorrect type of rule: {specification_type}')


class ExclusionSpecification:
    """Verifies whether entities matches set of specification entries.

    Attributes:
        specifications: All specification entries within the specification.
    """

    def __init__(
        self,
        specifications: Sequence[Sequence[BaseExclusionSpecificationEntry]]
        | None = None
    ) -> None:
        """Initializes ExclusionSpecification.

        Args:
            specifications: All specification entries within the specification.
        Returns:
        """
        self.specifications = specifications

    def apply_specifications(
            self,
            placements: GaarfReport,
            include_reason: bool = True,
            include_matching_placement: bool = True) -> GaarfReport | None:
        """Get placements that satisfy exclusion specifications entries.

        Args:
            placements:
                Report to be checked against specifications.
            include_reason:
                Whether to include exclusion reason to output report.
            include_matching_placement:
                Whether to include matching placement to output report.
        Returns:
            Report filtered to placements that matches all
            specification entries.
        """
        if not self.specifications:
            return placements
        to_be_excluded_placements = []
        extra_columns: list[str] = []
        if include_reason:
            extra_columns.append('reason')
        if include_matching_placement:
            extra_columns.append('matching_placement')
        for placement in placements:
            reason, matching_placement = self.satisfies(placement)
            extra_data: list = []
            if reason:
                if include_reason:
                    reason_str = ','.join(list(itertools.chain(*reason)))
                    extra_data = [reason_str]
                    if include_matching_placement:
                        extra_data.append(matching_placement)
                to_be_excluded_placements.append(placement.data + extra_data)
        if to_be_excluded_placements:
            return GaarfReport(
                results=to_be_excluded_placements,
                column_names=placements.column_names + extra_columns)
        return None

    def satisfies(self, placement: GaarfRow) -> tuple[list[str], dict]:
        """Verity whether a single entity satisfies the specifications.

        Args:
            placement: GaarfRow object with placement data.
        Returns:
            Tuple with list of rules that placement satisfies and
            placement itself.
        """
        rules_satisfied: list[str] = []
        placement_satisfied: dict = {}
        for spec_entry in self.specifications:
            spec_satisfied: list[str] = []
            for spec in spec_entry:
                if spec.exclusion_type != enums.ExclusionTypeEnum.GOOGLE_ADS_INFO:
                    formatted_exclusion_type = spec.exclusion_type.name.replace(
                        '_INFO', '')
                    if formatted_exclusion_type != placement.placement_type:
                        continue
                is_satisfied, placement_satisfied = spec.is_satisfied_by(
                    placement)
                if is_satisfied:
                    spec_satisfied.append(str(spec))
                    continue
            if len(spec_satisfied) == len(spec_entry):
                rules_satisfied.append(spec_satisfied)
        return rules_satisfied, placement_satisfied

    @property
    def ads_specs_entries(self) -> ExclusionSpecification:
        """Specification filtered to Ads specific specification entries."""
        return self._get_specification_subset(include_exclusion_types={
            enums.ExclusionTypeEnum.GOOGLE_ADS_INFO,
        })

    @property
    def non_ads_specs_entries(self) -> ExclusionSpecification:
        """Specification filtered to non-Ads specific specification entries."""
        return self._get_specification_subset(exclude_exclusion_types={
            enums.ExclusionTypeEnum.GOOGLE_ADS_INFO,
        })

    @property
    def parsable_spec_entries(self) -> ExclusionSpecification:
        """Specification filtered to parsable specification entries."""
        return self._get_specification_subset(
            include_exclusion_types=_PARSEABLE_PLACEMENT_TYPES_ENUMS)

    def _get_specification_subset(
        self,
        include_exclusion_types: set[enums.ExclusionTypeEnum] | None = None,
        exclude_exclusion_types: set[enums.ExclusionTypeEnum] | None = None
    ) -> ExclusionSpecification:
        """Helper method for building new specification from subset.

        Args:
            include_exclusion_types:
                Set of exclusion to include into new specifications.
            exclude_exclusion_types:
                Set of exclusion to include from new specifications.
        Returns:
            New ExclusionSpecification.
        """
        if not self.specifications:
            return ExclusionSpecification()
        specifications: list[list[BaseExclusionSpecificationEntry]] = []
        if include_exclusion_types:
            allowed_exclusion_types = include_exclusion_types
        else:
            exclude_exclusion_types = exclude_exclusion_types or set()
            allowed_exclusion_types = set(
                enums.ExclusionTypeEnum).difference(exclude_exclusion_types)
        for specification in self.specifications:
            matching_specification = []
            for specification_entry in specification:
                if (specification_entry.exclusion_type
                        in allowed_exclusion_types):
                    matching_specification.append(specification_entry)
            if matching_specification:
                specifications.append(matching_specification)
        return ExclusionSpecification(specifications=specifications)

    def define_runtime_options(self) -> RuntimeOptions:
        """Generates runtime options based on provided specifications."""
        if not self.specifications:
            return RuntimeOptions()
        is_regular_query = False
        is_conversion_query = False
        conversion_name = ''
        conversion_rules: list[BaseExclusionSpecificationEntry] = []
        for specification in self.specifications:
            for rule in specification:
                if rule.name == 'conversion_name':
                    is_conversion_query = True
                    conversion_name = rule.value
                    conversion_rules.append(rule)
                else:
                    is_regular_query = True
        return RuntimeOptions(
            is_regular_query=is_regular_query,
            is_conversion_query=is_conversion_query,
            conversion_name=conversion_name,
            conversion_rules=conversion_rules)

    @classmethod
    def from_rules(
            cls, parsed_rules: list[list[rules_parser.Rule]] | None
    ) -> ExclusionSpecification:
        """Convert Exclusion rules into specifications."""
        if not parsed_rules:
            return ExclusionSpecification()
        specifications = []
        for rules in parsed_rules:
            specification_entry = []
            for rule in rules:
                specification_entry.append(
                    create_exclusion_specification_entry(
                        rule.exclusion_type, rule.exclusion_rule))
            specifications.append(specification_entry)
        return ExclusionSpecification(specifications)

    @classmethod
    def from_expression(cls, rule_expression: str) -> ExclusionSpecification:
        """Convert raw string rules into specifications."""
        return ExclusionSpecification.from_rules(
            rules_parser.generate_rules(rule_expression))

    def __eq__(self, other) -> bool:
        return self.specifications == other.specifications

    def __bool__(self) -> bool:
        return bool(self.specifications)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Callable
from typing import Type

from googleads_housekeeper.domain import commands
from googleads_housekeeper.domain import events
from googleads_housekeeper.services import unit_of_work

Message = commands.Command | events.Event


class MessageBus:

    def __init__(self, uow: unit_of_work.AbstractUnitOfWork,
                 event_handlers: dict[Type[events.Event], Callable],
                 command_handlers: dict[Type[commands.Command],
                                        Callable], dependencies: dict) -> None:
        self.uow = uow
        self.event_handlers = event_handlers
        self.command_handlers = command_handlers
        self.dependencies = dependencies
        self.return_value = None
        self.queue = []

    def handle(self, message: Message):
        self.queue = [message]
        while self.queue:
            message = self.queue.pop(0)
            if isinstance(message, events.Event):
                self.handle_event(message)
            elif isinstance(message, commands.Command):
                self.handle_command(message)

        if self.return_value is not None:
            return self.return_value
        return None

    def handle_command(self, command: commands.Command) -> None:
        handler = self.command_handlers[type(command)]
        result = handler(command)
        self.return_value = result
        self.queue.extend(self.uow.collect_new_events())

    def handle_event(self, event: events.Event) -> None:
        for handler in self.event_handlers[type(event)]:
            handler(event)
            self.queue.extend(self.uow.collect_new_events())



from __future__ import annotations

import gaarf
import pytest

from googleads_housekeeper.domain.placement_handler import placement_excluder
from googleads_housekeeper.services import enums

_TEST_CUSTOMER_ID = 123456789
_TEST_CAMPAIGN_ID = 23456
_TEST_AD_GROUP_ID = 34567
_TEST_CRITERION_ID = 456789
_REPORT_COLUMN_NAMES = [
    'customer_id', 'campaign_id', 'campaign_type', 'ad_group_id',
    'placement_type', 'placement', 'criterion_id'
]
_REPORT_COLUMN_NAMES_WITH_ALLOWLISTING = _REPORT_COLUMN_NAMES + [
    'allowlisted', 'excluded_already'
]


class TestPlacementExcluder:

    @pytest.fixture
    def test_client(self, mocker):
        mocker.patch('google.ads.googleads.client.oauth2', return_value=[])
        return gaarf.api_clients.GoogleAdsApiClient()

    @pytest.fixture
    def excluder(self, test_client):
        return placement_excluder.PlacementExcluder(
            client=test_client,
            exclusion_level=enums.ExclusionLevelEnum.AD_GROUP)

    @pytest.fixture
    def placement_exclusion_lists(self):
        return {
            f'CPR Negative placements list - Campaign: {_TEST_CAMPAIGN_ID}':
                f'customers/{_TEST_CUSTOMER_ID}/sharedSet/123'
        }

    @pytest.fixture
    def website_placement_display_campaign(self):
        return gaarf.report.GaarfRow(
            data=[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'DISPLAY',
                _TEST_AD_GROUP_ID, 'WEBSITE', 'example.com',
                f'{_TEST_CRITERION_ID}'
            ],
            column_names=_REPORT_COLUMN_NAMES)

    @pytest.fixture
    def youtube_placement_video_campaign(self):
        return gaarf.report.GaarfRow(
            data=[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'VIDEO',
                _TEST_AD_GROUP_ID, 'YOUTUBE_VIDEO', 'test-video-id',
                f'{_TEST_CRITERION_ID}'
            ],
            column_names=_REPORT_COLUMN_NAMES)

    def test_create_placement_operation_returns_correct_ad_group_path(
            self, website_placement_display_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            website_placement_display_campaign, placement_exclusion_lists)

        expected_path = f'customers/{_TEST_CUSTOMER_ID}/adGroups/{_TEST_AD_GROUP_ID}'
        assert operation.exclusion_operation.create.ad_group == expected_path
        assert 'example.com' in operation.exclusion_operation.create.placement.url
        assert not operation.is_associatable
        assert not operation.shared_set_resource_name
        assert hasattr(operation.exclusion_operation.create, 'negative')

    def test_create_placement_operation_returns_correct_campaign_path(
            self, website_placement_display_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.CAMPAIGN
        excluder.exclusion_level = exclusion_level
        operation, _ = excluder._create_placement_operation(
            website_placement_display_campaign, placement_exclusion_lists)

        expected_path = f'customers/{_TEST_CUSTOMER_ID}/campaigns/{_TEST_CAMPAIGN_ID}'
        assert operation.exclusion_operation.create.campaign == expected_path
        assert hasattr(operation.exclusion_operation.create, 'negative')

    def test_create_placement_operation_returns_correct_customer_path(
            self, website_placement_display_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.ACCOUNT
        excluder.exclusion_level = exclusion_level
        operation, _ = excluder._create_placement_operation(
            website_placement_display_campaign, placement_exclusion_lists)

        assert not hasattr(operation.exclusion_operation.create, 'negative')

    def test_create_placement_operation_returns_correct_placement_type(
            self, website_placement_display_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            website_placement_display_campaign, placement_exclusion_lists)

        assert 'example.com' in operation.exclusion_operation.create.placement.url

    def test_create_placement_operation_returns_correct_placement_type_for_mobile_application(
            self, excluder, placement_exclusion_lists):

        placement_info = gaarf.report.GaarfRow(
            data=[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'DISPLAY',
                _TEST_AD_GROUP_ID, 'MOBILE_APPLICATION',
                'com.example.googleplay', '_TEST_CRITERION_ID'
            ],
            column_names=_REPORT_COLUMN_NAMES)

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            placement_info, placement_exclusion_lists)

        assert 'com.example.googleplay' in (
            operation.exclusion_operation.create.mobile_application.app_id)

    def test_create_placement_operation_returns_correct_placement_type_for_youtube_video(
            self, excluder, placement_exclusion_lists):

        placement_info = gaarf.report.GaarfRow(
            data=[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'DISPLAY',
                _TEST_AD_GROUP_ID, 'YOUTUBE_VIDEO', 'test-video-id',
                '_TEST_CRITERION_ID'
            ],
            column_names=_REPORT_COLUMN_NAMES)

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            placement_info, placement_exclusion_lists)

        assert 'test-video-id' in (
            operation.exclusion_operation.create.youtube_video.video_id)

    def test_create_placement_operation_returns_correct_placement_type_for_youtube_channel(
            self, excluder, placement_exclusion_lists):

        placement_info = gaarf.report.GaarfRow(
            data=[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'DISPLAY',
                _TEST_AD_GROUP_ID, 'YOUTUBE_CHANNEL', 'test-channel-id',
                '_TEST_CRITERION_ID'
            ],
            column_names=_REPORT_COLUMN_NAMES)

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            placement_info, placement_exclusion_lists)

        assert 'test-channel-id' in (
            operation.exclusion_operation.create.youtube_channel.channel_id)

    def test_create_placement_operation_does_not_create_shared_for_website(
            self, website_placement_display_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            website_placement_display_campaign, placement_exclusion_lists)

        assert not operation.shared_set_resource_name

    def test_create_placement_operation_returns_shared_set_operation_for_video_campaign(
            self, youtube_placement_video_campaign, excluder,
            placement_exclusion_lists):

        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        operation, _ = excluder._create_placement_operation(
            youtube_placement_video_campaign, placement_exclusion_lists)

        assert operation.exclusion_operation.create.shared_set in (
            placement_exclusion_lists.values())
        assert not operation.is_associatable

    def test_create_placement_exclusion_operations_returns_only_operations_for_display_campaign(
            self, excluder, placement_exclusion_lists):
        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        report = gaarf.report.GaarfReport(
            results=[[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'DISPLAY',
                _TEST_AD_GROUP_ID, 'WEBSITE', 'example.com',
                '_TEST_CRITERION_ID', False, False
            ]],
            column_names=_REPORT_COLUMN_NAMES_WITH_ALLOWLISTING)
        exclusion_operations = excluder._create_exclusion_operations(
            report, placement_exclusion_lists)

        assert _TEST_CUSTOMER_ID in exclusion_operations.placement_exclusion_operations
        assert not exclusion_operations.shared_set_creation_operations
        assert not exclusion_operations.campaign_set_association_operations

    def test_create_placement_exclusion_operations_returns_nothing_for_allowlisted_placement(
            self, excluder, placement_exclusion_lists):
        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        report = gaarf.report.GaarfReport(
            results=[[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'VIDEO',
                _TEST_AD_GROUP_ID, 'YOUTUBE_VIDEO', 'test-video-id',
                '_TEST_CRITERION_ID', True, False
            ]],
            column_names=_REPORT_COLUMN_NAMES_WITH_ALLOWLISTING)
        exclusion_operations = excluder._create_exclusion_operations(
            report, placement_exclusion_lists)
        assert not exclusion_operations.placement_exclusion_operations
        assert not exclusion_operations.shared_set_creation_operations
        assert not exclusion_operations.campaign_set_association_operations

    def test_create_placement_exclusion_operations_returns_only_share_set_operations_for_video_campaign(
            self, excluder, placement_exclusion_lists):
        exclusion_level = enums.ExclusionLevelEnum.AD_GROUP
        report = gaarf.report.GaarfReport(
            results=[[
                _TEST_CUSTOMER_ID, _TEST_CAMPAIGN_ID, 'VIDEO',
                _TEST_AD_GROUP_ID, 'YOUTUBE_VIDEO', 'test-video-id',
                '_TEST_CRITERION_ID', False, False
            ]],
            column_names=_REPORT_COLUMN_NAMES_WITH_ALLOWLISTING)
        exclusion_operations = excluder._create_exclusion_operations(
            report, placement_exclusion_lists)

        assert not exclusion_operations.placement_exclusion_operations
        assert _TEST_CUSTOMER_ID in exclusion_operations.shared_set_creation_operations
        assert not exclusion_operations.campaign_set_association_operations



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining configuration entities."""
from __future__ import annotations

import dataclasses
import uuid


@dataclasses.dataclass
class Config:
    mcc_id: str
    email_address: str
    always_fetch_youtube_preview_mode: bool = True
    save_to_db: bool = True
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


@dataclasses.dataclass
class CustomerIds:
    mcc_id: str
    account_name: str
    id: str


@dataclasses.dataclass
class MccIds:
    id: str
    account_name: str



from __future__ import annotations

import pytest

from googleads_housekeeper.domain.placement_handler import entities
from googleads_housekeeper.services import enums


class TestPlacementEntity:

    def test_placements_entity_empty_placement_types_returns_all_default_placement_types(
            self):
        placements = entities.Placements(placement_types=None)
        expected = '","'.join(placements._PLACEMENT_TYPES)
        assert placements.placement_types == expected

    def test_placements_entity_returns_formatted_placement_types_from_tuple(self):
        placements = entities.Placements(
            placement_types=('WEBSITE', 'MOBILE_APPLICATION'))
        assert placements.placement_types == 'WEBSITE","MOBILE_APPLICATION'

    def test_placements_entity_returns_formatted_placement_types_from_string(self):
        placements = entities.Placements(
            placement_types='WEBSITE,MOBILE_APPLICATION')
        assert placements.placement_types == 'WEBSITE","MOBILE_APPLICATION'

    def test_placements_entity_wrong_placement_type_raises_value_error(self):
        with pytest.raises(ValueError):
            entities.Placements(placement_types=('WRONG_PLACEMENT',))

    def test_placements_entity_wrong_start_date_return_value_error(self):
        with pytest.raises(ValueError):
            entities.Placements(start_date='1/1/1/')

    def test_placements_entity_wrong_end_date_return_value_error(self):
        with pytest.raises(ValueError):
            entities.Placements(end_date='1/1/1/')

    def test_placements_entity_start_date_greater_than_end_date_returns_value_error(
            self):
        with pytest.raises(ValueError):
            entities.Placements(start_date='2023-01-01', end_date='2022-01-01')

    def test_placements_entity_empty_campaign_types_returns_all_default_campaign_types(
            self):
        placements = entities.Placements(campaign_types=None)
        expected = '","'.join(placements._CAMPAIGN_TYPES)
        assert placements.campaign_types == expected

    def test_placements_entity_returns_formatted_campaign_types_from_tuple(self):
        placements = entities.Placements(campaign_types=('DISPLAY', 'VIDEO'))
        assert placements.campaign_types == 'DISPLAY","VIDEO'

    def test_placements_entity_returns_formatted_campaign_types_from_string(self):
        placements = entities.Placements(campaign_types='DISPLAY,VIDEO')
        assert placements.campaign_types == 'DISPLAY","VIDEO'

    def test_placements_entity_wrong_campaign_type_raises_value_error(self):
        with pytest.raises(ValueError):
            entities.Placements(campaign_types=('WRONG_CAMPAIGN_TYPE',))

    def test_placements_entity_wrong_granularity_raises_value_error(self):
        with pytest.raises(ValueError):
            entities.Placements(
                placement_level_granularity='wrong_placement_view')


class TestPlacementConversionSplitEntity:

    def test_placements_conversion_split_is_subclass_of_placement_entity(self):
        placements_conversion_split = entities.PlacementsConversionSplit(
            placement_types=None)
        assert isinstance(placements_conversion_split, entities.Placements)


class TestAlreadyExcludedPlacements:

    @pytest.mark.parametrize('exclusion_level,expected_resource', [
        (enums.ExclusionLevelEnum.ACCOUNT, 'customer_negative_criterion'),
        (enums.ExclusionLevelEnum.CAMPAIGN, 'campaign_criterion'),
        (enums.ExclusionLevelEnum.AD_GROUP, 'ad_group_criterion'),
    ])
    def test_already_excluded_placements_query_is_correct(
            self, exclusion_level, expected_resource):
        already_excluded_placements = entities.AlreadyExcludedPlacements(
            exclusion_level)
        assert expected_resource in already_excluded_placements.query_text



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import functools
import operator

import gaarf

from googleads_housekeeper import views
from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.domain.core import task
from googleads_housekeeper.domain.external_parsers import external_entity_parser
from googleads_housekeeper.domain.placement_handler import placement_excluder
from googleads_housekeeper.domain.placement_handler import placement_fetcher
from googleads_housekeeper.domain.placement_handler import placement_info_extractor
from googleads_housekeeper.services import enums
from googleads_housekeeper.services import unit_of_work


@dataclasses.dataclass
class RuntimeOptions:
    """Options for configuring fetching and saving of data."""
    always_fetch_youtube_preview_mode: bool = True
    save_to_db: bool = True


def exclude_placements(
        task_obj: task.Task, uow: unit_of_work.AbstractUnitOfWork, client,
        runtime_options: RuntimeOptions) -> placement_excluder.ExclusionResult:
    """Excludes data from Google Ads based on specified criteria.

    Args:
        task_obj: A Task for identifying placements suitable for exclusion.
        uow: Unit of Work to handle the transaction.
        runtime_options: Options for configuring fetching and saving of data.
    """
    report_fetcher = gaarf.query_executor.AdsReportFetcher(client)
    fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
    excluder = placement_excluder.PlacementExcluder(client,
                                                    task_obj.exclusion_level)
    to_be_excluded_placements = find_placements_for_exclusion(
        task_obj, uow, report_fetcher, runtime_options)
    customer_ids = to_be_excluded_placements['customer_id'].to_list(
        flatten=True, distinct=True)
    placement_exclusion_lists = fetcher.get_placement_exclusion_lists(
        customer_ids)
    return excluder.exclude_placements(to_be_excluded_placements,
                                       placement_exclusion_lists)


def _is_already_excluded_placement(
        placement_info: gaarf.report.GaarfRow,
        already_excluded_placements: dict[int, list[int]],
        exclusion_level: enums.ExclusionLevelEnum) -> bool:
    """Verifies that a placement has already been added to exclusions.

    Args:
        placement_info:
            GaarfRow containing information on a single placement.
        already_excluded_placements:
            Mapping for placements added as negative criteria on a
            specific level (ACCOUNT, CAMPAIGN, AD_GROUP).
        exclusion_level:
            Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).
    Returns:
        Description of return.
    Raises:
        ValueError: When incorrect exclusion level is specified.
    """
    if exclusion_level == enums.ExclusionLevelEnum.ACCOUNT:
        level_id = placement_info.account_id
    elif exclusion_level == enums.ExclusionLevelEnum.CAMPAIGN:
        level_id = placement_info.campaign_id
    elif exclusion_level == enums.ExclusionLevelEnum.AD_GROUP:
        level_id = placement_info.ad_group_id
    else:
        raise ValueError(f'Incorrect exclusion_level: {exclusion_level}')
    return (placement_info.placement
            in already_excluded_placements.get(level_id, []))


def find_placements_for_exclusion(
    task_obj: task.Task,
    uow: unit_of_work.AbstractUnitOfWork,
    report_fetcher: gaarf.query_executor.AdsReportFetcher | None = None,
    runtime_options: RuntimeOptions = RuntimeOptions(),
) -> gaarf.report.GaarfReport | None:
    """Identifies the placements suitable for exclusion.

    Args:
        task_obj: A Task for identifying placements suitable for exclusion.
        uow: Unit of Work to handle the transaction.
        report_fetcher: An instance of AdsReportFetcher to handle API requests.
        runtime_options: Options for configuring fetching and saving of data.
    Returns:
        Report containing placements suitable for exclusion.
    """
    fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
    extractor = placement_info_extractor.PlacementInfoExtractor(uow)
    external_parser = external_entity_parser.ExternalEntitiesParser(uow)
    specification = (
        exclusion_specification.ExclusionSpecification.from_expression(
            task_obj.exclusion_rule))
    reports: list[gaarf.report.GaarfReport] = []
    for account in task_obj.accounts:
        account_allowlisted_placements = views.allowlisted_placements_as_tuples(
            uow, account)
        if not (placements := fetcher.get_placements_for_account(
                account, task_obj)):
            continue
        already_excluded_placements = fetcher.get_already_excluded_placements(
            account, task_obj.exclusion_level)
        if ads_specs := specification.ads_specs_entries:
            placements = ads_specs.apply_specifications(
                placements,
                include_reason=False,
                include_matching_placement=False)
            if not placements:
                continue
        if runtime_options.always_fetch_youtube_preview_mode:
            youtube_channel_specification = (
                exclusion_specification.ExclusionSpecification(
                    specifications=[[
                        exclusion_specification
                        .YouTubeChannelExclusionSpecificationEntry(
                            'subscriberCount > 0')
                    ],
                                    [
                                        exclusion_specification.
                                        YouTubeVideoExclusionSpecificationEntry(
                                            'likesCount > 0')
                                    ]]))
            external_parser.parse_specification_chain(
                placements, youtube_channel_specification)
        if non_ads_specs := specification.non_ads_specs_entries:
            external_parser.parse_specification_chain(placements, non_ads_specs)
        placements = inject_extra_data(placements, extractor,
                                       account_allowlisted_placements,
                                       already_excluded_placements,
                                       task_obj.exclusion_level)
        if (placements := specification.apply_specifications(placements)):
            reports.append(placements)
    if reports:
        return functools.reduce(operator.add, reports)
    return None


def inject_extra_data(
        placements: gaarf.report.GaarfReport,
        extractor: placement_info_extractor.PlacementInfoExtractor,
        allowlisted_placements: list[tuple[str, str, str]],
        already_excluded_placements: dict[int, list[int]],
        exclusion_level: enums.ExclusionLevelEnum) -> gaarf.report.GaarfReport:
    """Add extra placement information to the placement report.

    Args:
        placements:
            Report containing placement data.
        extractor:
            Responsible for getting parsed placement metadata data from DB.
        allowlisted_placements:
            Info on placements that shouldn't be excluded form a given account.
        already_excluded_placements:
            Mapping between entity_id from specified exclusion level to all
            negative criteria ids.
        exclusion_level: Level of exclusion (ACCOUNT, CAMPAIGN, AD_GROUP).
    Returns:
        Placement data report with new columns
        ('allowlisted', 'already_excluded', 'extra_info').
    """
    has_allowlisted_placements = allowlisted_placements or False
    for placement in placements:
        if has_allowlisted_placements:
            placement['allowlisted'] = (
                placement.name, placement.placement_type,
                placement.customer_id) in allowlisted_placements
        else:
            placement['allowlisted'] = False
        placement['excluded_already'] = _is_already_excluded_placement(
            placement, already_excluded_placements, exclusion_level)
        placement['extra_info'] = extractor.extract_placement_info(placement)
    return placements



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses
import datetime
import enum
import math
import uuid

from croniter import croniter

from googleads_housekeeper.services import enums


class TaskStatus(enum.Enum):
    """An enumeration representing the status of a task.

    Attributes:
        ACTIVE (int): Represents an active task.
        INACTIVE (int): Represents an inactive task.
    """
    ACTIVE = 0
    INACTIVE = 1


class TaskOutput(enum.Enum):
    """An enumeration representing the artifact/output options for a task.

    Attributes:
        NOTIFY (int): Output option to notify about the task.
        EXCLUDE (int): Output option to exclude the task.
        EXCLUDE_AND_NOTIFY (int): Output option to exclude the task and notify about it.
    """
    NOTIFY = 1
    EXCLUDE = 2
    EXCLUDE_AND_NOTIFY = 3


@dataclasses.dataclass
class Task:
    """Defines task that needs to be run by an application

    Task contains all necessary information to find the appropriate placements
    that should be excluded and perform the exclusion.

    Attributes:
        name: Task name.
        exclusion_rule: String version of exclusion rule.
        date_range: Lookback days for data fetching from Google Ads API.
        from_days_ago: start_date of data fetching from Google Ads API.
        exclusion_level: Level of exclusion (AD_GROUP, CAMPAIGN, etc.).
        output: Desired action (EXCLUDE, NOTIFY, etc.)
        status: Task status.
        schedule: Optional task schedule.
        creation_date: Time when task was created.
        id: Unique task identifier.
    """
    name: str
    exclusion_rule: str
    customer_ids: str
    date_range: int = 7
    from_days_ago: int = 0
    exclusion_level: enums.ExclusionLevelEnum = enums.ExclusionLevelEnum.AD_GROUP
    placement_types: str | None = None
    output: TaskOutput = TaskOutput.EXCLUDE_AND_NOTIFY
    status: TaskStatus = TaskStatus.ACTIVE
    schedule: str | None = None
    creation_date: datetime.datetime = datetime.datetime.now()
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self) -> None:
        """Ensures safe casting to proper enums."""
        self.output = self._cast_to_enum(TaskOutput, self.output)
        self.status = self._cast_to_enum(TaskStatus, self.status)
        self.exclusion_level = self._cast_to_enum(enums.ExclusionLevelEnum,
                                                  self.exclusion_level)

    def to_dict(self) -> dict:
        """Convert class to dictionary.

        Enum attributes of the class are represented as names.
        """
        task_dict = {}
        for key, value in dataclasses.asdict(self).items():
            if hasattr(value, 'value'):
                task_dict[key] = value.name
            else:
                task_dict[key] = value
        return task_dict

    @property
    def cron_schedule(self) -> str | None:
        """Builds cron schedule based on creation time and schedule."""
        minute = self.creation_date.strftime('%M')
        hour = self.creation_date.strftime('%H')

        if not (schedule := self.schedule) or self.schedule == '0':
            return None
        if 0 < int(schedule) < 24:
            schedule = f'{minute} */{schedule} * * *'
        elif int(schedule) >= 24:
            days = math.floor(int(self.schedule) / 24)
            schedule = f'{minute} {hour} */{days} * *'
        return schedule

    @property
    def start_date(self) -> str:
        """Returns formatted start_date."""
        return (datetime.datetime.now() - datetime.timedelta(
            days=int(self.date_range))).date().strftime('%Y-%m-%d')

    @property
    def end_date(self) -> str:
        """Returns formatted end_date."""
        return (datetime.datetime.now() - datetime.timedelta(
            days=int(self.from_days_ago))).date().strftime('%Y-%m-%d')

    @property
    def accounts(self) -> list[str]:
        """Returns formatted account list."""
        return self.customer_ids.split(',') if isinstance(
            self.customer_ids, str) else self.customer_ids

    @property
    def next_run(self) -> str:
        """Returns next run as a cron expression."""
        if not (schedule := self.cron_schedule):
            return 'Not scheduled'
        return croniter(schedule, datetime.datetime.now()).get_next(
            datetime.datetime).strftime('%Y-%m-%d %H:%M')

    def _cast_to_enum(self, enum_: type[enum.Enum],
                      value: str | enum.Enum) -> enum.Enum:
        return enum_[value] if isinstance(value, str) else value



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from collections import abc
from dataclasses import dataclass

from googleads_housekeeper.domain.core import execution
from googleads_housekeeper.domain.core import task


class Command:
    ...


@dataclass
class RunTask(Command):
    id: int
    type: execution.ExecutionTypeEnum
    exclusion_rule: str = ''
    save_to_db: bool = True


@dataclass
class SaveTask(Command):
    exclusion_rule: str
    customer_ids: str | list[str]
    from_days_ago: int = 0
    date_range: int = 7
    exclusion_level: str = 'AD_GROUP'
    output: str = task.TaskOutput.EXCLUDE_AND_NOTIFY.name
    name: str | None = None
    schedule: str | None = None
    placement_types: str | None = None
    task_id: str | None = None

    def __post_init__(self):
        self.customer_ids = ','.join(self.customer_ids) if isinstance(
            self.customer_ids, abc.MutableSequence) else self.customer_ids


@dataclass
class DeleteTask(Command):
    task_id: int


@dataclass
class RunManualExclusion(Command):
    customer_ids: str
    placements: list[list]
    header: list[str]
    exclusion_level: str


@dataclass
class PreviewPlacements(Command):
    exclusion_rule: str
    placement_types: tuple[str, ...] | None
    customer_ids: str | list[str]
    from_days_ago: int
    date_range: int
    exclusion_level: str = 'AD_GROUP'
    exclude_and_notify: str = 'EXCLUDE_AND_NOTIFY'
    save_to_db: bool = True

    def __post_init__(self):
        self.placement_types = tuple(
            self.placement_types.split(',')) if self.placement_types else None
        self.from_days_ago = int(self.from_days_ago)
        self.date_range = int(self.date_range)


@dataclass
class AddToAllowlisting(Command):
    type: str
    name: str
    account_id: str


@dataclass
class RemoveFromAllowlisting(Command):
    type: str
    name: str
    account_id: str


@dataclass
class SaveConfig(Command):
    id: str
    mcc_id: str
    email_address: str
    always_fetch_youtube_preview_mode: bool = True
    save_to_db: bool = False


@dataclass
class GetCustomerIds(Command):
    mcc_id: str


@dataclass
class GetMccIds(Command):
    root_mcc_id: int


@dataclass
class SaveWebsiteInfo(Command):
    website_info: dict


@dataclass
class SaveChannelInfo(Command):
    channel_info: dict


@dataclass
class SaveVideoInfo(Command):
    video_info: dict



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Application handlers."""
from __future__ import annotations

import json
import logging
from copy import copy
from dataclasses import asdict
from datetime import datetime
from typing import Any

from gaarf.api_clients import GoogleAdsApiClient
from gaarf.query_executor import AdsReportFetcher
from gaarf.report import GaarfReport

from googleads_housekeeper.adapters.notifications import MessagePayload
from googleads_housekeeper.adapters.publisher import BasePublisher
from googleads_housekeeper.domain import commands
from googleads_housekeeper.domain import events
from googleads_housekeeper.domain.core import execution
from googleads_housekeeper.domain.core import settings
from googleads_housekeeper.domain.core import task
from googleads_housekeeper.domain.placement_handler import entities as placement_entities
from googleads_housekeeper.domain.placement_handler import placement_excluder
from googleads_housekeeper.domain.placement_handler import placement_fetcher
from googleads_housekeeper.domain.placement_handler import placement_service
from googleads_housekeeper.services import unit_of_work


def get_mcc_accounts(
        cmd: commands.GetMccIds, uow: unit_of_work.AbstractUnitOfWork,
        ads_api_client: GoogleAdsApiClient) -> list[dict[str, str]]:
    """Fetches and updates mcc accounts in the repository.

    Args:
        cmd: command for getting MccIds.
        uow: unit of work to handle the transaction.
        ads_api_client: client to perform report fetching.
    Returns:
        All mcc accounts available under given root_mcc_id.
    """
    mccs = get_accessible_mccs(ads_api_client, cmd.root_mcc_id)
    result: list[dict[str, str]] = []
    with uow:
        saved_mcc_ids = {int(account.id) for account in uow.mcc_ids.list()}
        fetched_mcc_ids = set(mccs['account_id'].to_list(
            flatten=True, distinct=True))
        new_accounts = fetched_mcc_ids.difference(saved_mcc_ids)
        to_be_removed_accounts = saved_mcc_ids.difference(fetched_mcc_ids)
        for row in mccs:
            row_dict = {'id': row.account_id, 'account_name': row.account_name}
            result.append(row_dict)
            if row.account_id in new_accounts:
                uow.mcc_ids.add(settings.MccIds(**row_dict))
        for account_id in to_be_removed_accounts:
            uow.mcc_ids.delete(account_id)
        uow.commit()
    return result


def get_accessible_mccs(ads_api_client: GoogleAdsApiClient,
                        root_mcc_id: int) -> GaarfReport:
    """Fetches all available mccs under a root_mcc_id.

    Args:
        ads_api_client: client to perform report fetching.
        root_mcc_id: root mmc id for which leaf mccs should be fetched.
    Returns:
        Report with mcc names and ids."""
    query = """
        SELECT
            customer_client.descriptive_name AS account_name,
            customer_client.id AS account_id
        FROM customer_client
        WHERE customer_client.manager = TRUE
        AND customer_client.status = "ENABLED"
        """
    report_fetcher = AdsReportFetcher(api_client=ads_api_client)
    return report_fetcher.fetch(query, root_mcc_id)


def get_customer_ids(
        cmd: commands.GetCustomerIds, uow: unit_of_work.AbstractUnitOfWork,
        ads_api_client: GoogleAdsApiClient) -> list[dict[str, int]]:
    """Fetches accounts under a given mcc_id and update them in the repository.

    Args:
        cmd: command for getting customer ids under a given mcc_id.
        uow: unit of work to handle the transaction.
        ads_api_client: client to perform report fetching.
    Returns:
        All accounts available under given mcc_id.
    """
    customer_ids_query = """
    SELECT
        customer_client.descriptive_name AS account_name,
        customer_client.id AS account_id
    FROM customer_client
    WHERE customer_client.manager = FALSE
    AND customer_client.status = "ENABLED"
    """
    report_fetcher = AdsReportFetcher(api_client=ads_api_client)
    customer_ids = report_fetcher.fetch(
        customer_ids_query, customer_ids=cmd.mcc_id)
    saved_customer_ids = set([
        int(account.id) for account in uow.customer_ids.get_by_conditions(
            {'mcc_id': cmd.mcc_id})
    ])
    fetched_customer_ids = set(customer_ids['account_id'].to_list(
        flatten=True, distinct=True))
    new_accounts = fetched_customer_ids.difference(saved_customer_ids)
    to_be_removed_accounts = saved_customer_ids.difference(fetched_customer_ids)
    result: list[dict[str, int]] = []
    with uow:
        for row in customer_ids:
            result.append({
                'account_name': row.account_name,
                'id': row.account_id
            })
            if row.account_id in new_accounts:
                uow.customer_ids.add(
                    settings.CustomerIds(
                        mcc_id=cmd.mcc_id,
                        account_name=row.account_name,
                        id=str(row.account_id)))
        for account_id in to_be_removed_accounts:
            uow.customer_ids.delete(account_id)
        uow.commit()
    return result


def run_manual_exclusion_task(
        cmd: commands.RunManualExclusion, uow: unit_of_work.AbstractUnitOfWork,
        ads_api_client: GoogleAdsApiClient) -> dict[str, int]:
    with uow:
        report_fetcher = AdsReportFetcher(ads_api_client)
        excluder = placement_excluder.PlacementExcluder(ads_api_client,
                                                        cmd.exclusion_level)
        fetcher = placement_fetcher.PlacementFetcher(report_fetcher)
        placement_exclusion_lists = fetcher.get_placement_exclusion_lists(
            cmd.customer_ids)
        exclusion_result = excluder.exclude_placements(
            to_be_excluded_placements=GaarfReport(
                results=cmd.placements, column_names=cmd.header),
            placement_exclusion_lists=placement_exclusion_lists)
        n_excluded_placements = len(exclusion_result.excluded_placements or [])
        n_associated_with_list_placements = len(
            exclusion_result.associated_with_list_placements or [])
        return {
            'excluded_placements': n_excluded_placements,
            'associated_with_list_placements': n_associated_with_list_placements
        }


def task_created(event: events.TaskCreated, publisher: BasePublisher) -> None:
    publisher.publish('task_created', event)


def task_with_schedule_created(event: events.TaskWithScheduleCreated,
                               publisher: BasePublisher) -> None:
    publisher.publish('task_created', event)


def task_updated(event: events.TaskUpdated, publisher: BasePublisher) -> None:
    publisher.publish('task_updated', event)


def task_schedule_updated(event: events.TaskScheduleUpdated,
                          publisher: BasePublisher) -> None:
    publisher.publish('task_updated', event)


def task_deleted(event: events.TaskDeleted, publisher: BasePublisher) -> None:
    publisher.publish('task_deleted', event)


def task_schedule_deleted(event: events.TaskScheduleDeleted,
                          publisher: BasePublisher) -> None:
    publisher.publish('task_deleted', event)


def run_task(cmd: commands.RunTask,
             uow: unit_of_work.AbstractUnitOfWork,
             ads_api_client: GoogleAdsApiClient,
             publisher: BasePublisher | None = None,
             save_to_db: bool = True) -> tuple[dict, MessagePayload | None]:
    with copy(uow) as uow:
        settings = uow.settings.list()
        task_obj = uow.tasks.get(task_id=cmd.id)
        start_time = datetime.now()
        exclusion_result = placement_service.exclude_placements(
            task_obj=task_obj,
            uow=uow,
            client=ads_api_client,
            runtime_options=placement_service.RuntimeOptions(
                always_fetch_youtube_preview_mode=False))
        end_time = datetime.now()
        n_excluded_placements = len(exclusion_result.excluded_placements or [])
        n_associated_with_list_placements = len(
            exclusion_result.associated_with_list_placements or [])
        execution_obj = execution.Execution(
            task=cmd.id,
            start_time=start_time,
            end_time=end_time,
            placements_excluded=n_excluded_placements,
            type=cmd.type)
        uow.executions.add(execution_obj)
        if save_to_db and exclusion_result.excluded_placements:
            for placement in exclusion_result.excluded_placements:
                if hasattr(placement, 'reason'):
                    exclusion_reason = placement.reason
                else:
                    exclusion_reason = ''
                uow.execution_details.add(
                    execution.ExecutionDetails(
                        execution_id=execution_obj.id,
                        placement=placement.name,
                        placement_type=placement.placement_type,
                        reason=exclusion_reason))
        uow.commit()
        if task_obj.output in (task.TaskOutput.NOTIFY,
                               task.TaskOutput.EXCLUDE_AND_NOTIFY):
            publisher.publish('task_run', execution_obj)
            if excluded_placements := exclusion_result.excluded_placements:
                message_payload = MessagePayload(
                    task_id=task_obj.id,
                    task_name=task_obj.name,
                    task_formula=task_obj.exclusion_rule,
                    placements_excluded_sample=excluded_placements[0:10],
                    total_placement_excluded=len(excluded_placements),
                    recipient=settings[0].email_address)
            else:
                message_payload = MessagePayload(
                    task_id=task_obj.id,
                    task_name=task_obj.name,
                    task_formula=task_obj.exclusion_rule,
                    placements_excluded_sample=None,
                    total_placement_excluded=0,
                    recipient=settings[0].email_address)
        else:
            message_payload = None
        return {
            'excluded_placements': n_excluded_placements,
            'associated_with_list_placements': n_associated_with_list_placements
        }, message_payload


def preview_placements(cmd: commands.PreviewPlacements,
                       uow: unit_of_work.AbstractUnitOfWork,
                       ads_api_client: GoogleAdsApiClient,
                       always_fetch_youtube_preview_mode: bool = True,
                       save_to_db: bool = True) -> dict[str, Any]:
    report_fetcher = AdsReportFetcher(api_client=ads_api_client)
    task_obj = task.Task(
        name='',
        exclusion_rule=cmd.exclusion_rule,
        customer_ids=cmd.customer_ids,
        date_range=cmd.date_range,
        from_days_ago=cmd.from_days_ago,
        exclusion_level=cmd.exclusion_level,
        placement_types=cmd.placement_types)
    to_be_excluded_placements = placement_service.find_placements_for_exclusion(
        task_obj=task_obj,
        uow=uow,
        report_fetcher=report_fetcher,
        runtime_options=placement_service.RuntimeOptions(
            always_fetch_youtube_preview_mode=always_fetch_youtube_preview_mode,
            save_to_db=save_to_db))
    if not to_be_excluded_placements:
        data = {}
    else:
        data = json.loads(
            to_be_excluded_placements.to_pandas().to_json(orient='index'))
    return {
        'data': data,
        'dates': {
            'date_from': task_obj.start_date,
            'date_to': task_obj.end_date
        }
    }


def save_task(
    cmd: commands.SaveTask,
    uow: unit_of_work.AbstractUnitOfWork,
) -> str:
    with uow:
        task_id = None
        task_obj = None
        if hasattr(cmd, 'task_id') and cmd.task_id:
            task_id = cmd.task_id
            task_obj = uow.tasks.get(task_id=task_id)
            update_dict = asdict(cmd)
            update_dict.pop('task_id')
            uow.tasks.update(task_obj.id, update_dict)
            uow.commit()
            uow.published_events.append(events.TaskUpdated(cmd.task_id))
            if cmd.schedule and cmd.schedule != task_obj.schedule:
                uow.published_events.append(
                    events.TaskScheduleUpdated(cmd.task_id, cmd.schedule))
            else:
                uow.published_events.append(
                    events.TaskScheduleDeleted(cmd.task_id))
        else:
            task_dict = asdict(cmd)
            task_dict.pop('task_id')
            task_obj = task.Task(**task_dict)
            uow.tasks.add(task_obj)
            uow.commit()
            task_id = task_obj.id
            uow.published_events.append(events.TaskCreated(task_id))
            if cmd.schedule:
                uow.published_events.append(
                    events.TaskWithScheduleCreated(task_id, cmd.name,
                                                   cmd.schedule))
        task_id = task_obj.id
        return str(task_id)


def delete_task(
    cmd: commands.DeleteTask,
    uow: unit_of_work.AbstractUnitOfWork,
):
    with copy(uow) as uow:
        task_obj = uow.tasks.get(task_id=cmd.task_id)
        if task_obj:
            uow.tasks.update(cmd.task_id, {'status': 'INACTIVE'})
            uow.commit()
            uow.published_events.append(events.TaskDeleted(cmd.task_id))
            uow.published_events.append(events.TaskScheduleDeleted(cmd.task_id))
        else:
            logging.warning('No task with id %d found!', cmd.id)


def save_config(
    cmd: commands.SaveConfig,
    uow: unit_of_work.AbstractUnitOfWork,
):
    with uow:
        if hasattr(cmd, 'id') and cmd.id:
            config_id = cmd.id
            config = uow.settings.get(config_id)
            update_dict = asdict(cmd)
            update_dict.pop('id')
            uow.settings.update(config.id, update_dict)
        else:
            config_dict = asdict(cmd)
            config_dict.pop('id')
            config = settings.Config(**config_dict)
            uow.settings.add(config)
        uow.commit()


def add_to_allowlisting(cmd: commands.AddToAllowlisting,
                        uow: unit_of_work.AbstractUnitOfWork) -> None:
    with uow:
        if not uow.allowlisting.get_by_conditions(asdict(cmd)):
            placement = placement_entities.AllowlistedPlacement(**asdict(cmd))
            uow.allowlisting.add(placement)
            uow.commit()


def remove_from_allowlisting(cmd: commands.RemoveFromAllowlisting,
                             uow: unit_of_work.AbstractUnitOfWork) -> None:
    with uow:
        if allowlisted_placement := uow.allowlisting.get_by_conditions(
                asdict(cmd)):
            uow.allowlisting.delete(allowlisted_placement[0].id)
            uow.commit()


def save_channel_info(cmd: commands.SaveChannelInfo,
                      uow: unit_of_work.AbstractUnitOfWork) -> None:
    with uow:
        uow.youtube_channel_info.add(cmd.channel_info)
        uow.commit()


def save_video_info(cmd: commands.SaveChannelInfo,
                    uow: unit_of_work.AbstractUnitOfWork) -> None:
    with uow:
        uow.youtube_video_info.add(cmd.video_info)
        uow.commit()


def save_website_info(cmd: commands.SaveChannelInfo,
                      uow: unit_of_work.AbstractUnitOfWork) -> None:
    with uow:
        uow.website_info.add(cmd.website_info)
        uow.commit()


EVENT_HANDLERS = {
    events.TaskWithScheduleCreated: [task_with_schedule_created],
    events.TaskScheduleUpdated: [task_schedule_updated],
    events.TaskScheduleDeleted: [task_schedule_deleted],
    events.TaskCreated: [task_created],
    events.TaskUpdated: [task_updated],
    events.TaskDeleted: [task_deleted]
}

COMMAND_HANDLERS = {
    commands.RunTask: run_task,
    commands.SaveTask: save_task,
    commands.DeleteTask: delete_task,
    commands.RunManualExclusion: run_manual_exclusion_task,
    commands.PreviewPlacements: preview_placements,
    commands.SaveConfig: save_config,
    commands.GetCustomerIds: get_customer_ids,
    commands.GetMccIds: get_mcc_accounts,
    commands.AddToAllowlisting: add_to_allowlisting,
    commands.RemoveFromAllowlisting: remove_from_allowlisting,
    commands.SaveChannelInfo: save_channel_info,
    commands.SaveVideoInfo: save_video_info,
    commands.SaveWebsiteInfo: save_website_info,
}



from __future__ import annotations

import gaarf
import pytest

from googleads_housekeeper.domain.external_parsers import website_parser
from googleads_housekeeper.domain.external_parsers import youtube_data_parser
from googleads_housekeeper.domain.placement_handler import placement_info_extractor


class TestPlacementInfoExtractor:

    @pytest.fixture
    def extractor(self, bus):
        return placement_info_extractor.PlacementInfoExtractor(bus.uow)

    def test_extract_placement_info_returns_website_info(self, extractor):
        gaarf_row = gaarf.report.GaarfRow(
            data=['example.com', 'WEBSITE'],
            column_names=['placement', 'placement_type'])
        expected_result = {
            'website_info':
                website_parser.WebsiteInfo(
                    placement='example.com',
                    title='fun games',
                    description='millions of fun games',
                    keywords='browser games, free games',
                    is_processed=True)
        }
        result = extractor.extract_placement_info(gaarf_row)
        assert result == expected_result

    def test_extract_placement_info_returns_youtube_channel_info(
            self, extractor):
        gaarf_row = gaarf.report.GaarfRow(
            data=['1kdjf0skdjfw0dsdf', 'YOUTUBE_CHANNEL'],
            column_names=['placement', 'placement_type'])
        expected_result = {
            'youtube_channel_info':
                youtube_data_parser.ChannelInfo(
                    placement='1kdjf0skdjfw0dsdf',
                    title='Gardening and Games',
                    description='Everything about Games and Gardens',
                    country='US',
                    viewCount=1000,
                    subscriberCount=100,
                    videoCount=10,
                    topicCategories='Gardening,Games',
                    is_processed=True)
        }
        result = extractor.extract_placement_info(gaarf_row)
        assert result == expected_result

    def test_extract_placement_info_returns_youtube_video_info(self, extractor):
        gaarf_row = gaarf.report.GaarfRow(
            data=['jojoh', 'YOUTUBE_VIDEO'],
            column_names=['placement', 'placement_type'])
        expected_result = {
            'youtube_video_info':
                youtube_data_parser.VideoInfo(
                    placement='jojoh',
                    title='Gardening and Games Vol. 2',
                    description=(
                        'The second volumes of the Gardening and Games series'),
                    defaultLanguage='en',
                    defaultAudioLanguage='en',
                    commentCount=10,
                    favouriteCount=10,
                    likeCount=10,
                    viewCount=1000,
                    madeForKids=True,
                    topicCategories='Gardening,Games',
                    tags='#multiplayer,#mro,#garden',
                    is_processed=True)
        }
        result = extractor.extract_placement_info(gaarf_row)
        assert result == expected_result

    def test_extract_placement_info_returns_emtpy_info(self, extractor):
        gaarf_row = gaarf.report.GaarfRow(
            data=['test-app', 'MOBILE_APPLICATION'],
            column_names=['placement', 'placement_type'])
        expected_result = {}
        result = extractor.extract_placement_info(gaarf_row)
        assert result == expected_result



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from googleads_housekeeper.adapters.legacy_adapter import TaskAdapter


def test_build_gads_filter_succeeds():
    task_adapter = TaskAdapter({
        'gads_filter': '(metrics.ctr > 2) OR '
                       '(metrics.video_views = 3 AND metrics.clicks > 20) OR '
                       '(metrics.conversions_from_interactions_rate > 20)'
    })

    actual_output = task_adapter._build_gads_filters()

    assert actual_output == (
        '(GOOGLE_ADS_INFO:ctr > 2) OR '
        '(GOOGLE_ADS_INFO:video_views = 3 AND GOOGLE_ADS_INFO:clicks > 20) OR '
        '(GOOGLE_ADS_INFO:conversions_from_interactions_rate > 20)')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for external entity parser."""
from __future__ import annotations

from gaarf.report import GaarfReport

from googleads_housekeeper.domain.core import exclusion_specification
from googleads_housekeeper.domain.external_parsers import (
    external_entity_parser)
from googleads_housekeeper.domain.external_parsers import website_parser


def test_external_entity_parser(mocker, bus):
    external_parser = external_entity_parser.ExternalEntitiesParser(bus.uow)
    specification_entry = (
        exclusion_specification.ContentExclusionSpecificationEntry(
            'placement regexp example'))
    specification = exclusion_specification.ExclusionSpecification(
        specifications=[[specification_entry]])
    report = GaarfReport(
        results=[['parsed-example.com', 'WEBSITE']],
        column_names=['placement', 'placement_type'])
    mocker.patch(
        'googleads_housekeeper.domain.external_parsers.website_parser.'
        'WebSiteParser.parse',
        return_value=[
            website_parser.WebsiteInfo(
                placement='parsed-example.com',
                title='example title',
                description='example description',
                keywords='example keywords',
                is_processed=True)
        ])

    external_parser.parse_specification_chain(
        entities=report, specification=specification.parsable_spec_entries)
    with bus.uow as uow:
        assert uow.website_info.get_by_conditions(
            {'placement': 'parsed-example.com'})



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for translating exclusion rule into specifications."""
from __future__ import annotations

import dataclasses
import re
from collections.abc import Sequence


@dataclasses.dataclass(frozen=True)
class Rule:
    """Holds information on exclusion rule."""
    exclusion_type: str
    exclusion_rule: str


def generate_rules(raw_rules: str | Sequence[str]) -> list[list[Rule]] | None:
    """Convert raw string rules into list of Rules.

    Rule as raw string can be expressed in an explicit
    (`exclusion_type:exclusion_rule`), implicit (`exclusion_rule`) format.

    Rules in each of these formats can be represented as sequence or a single
    string. And-rules are separated by `AND` keyword or comma (`,`),
    Or-rules are separated by `OR` keyword.

    If `exclusion_type` is not specified for a given rule it's treated
    as GOOGLE_ADS_INFO.
    """
    if not raw_rules:
        return None
    rules: list[list[Rule]] = []
    default_exclusion_type = 'GOOGLE_ADS_INFO'
    if isinstance(raw_rules, str):
        raw_rules = _format_raw_rules(raw_rules)
    for rule in raw_rules:
        current_exclusion_type = None
        if ',' in rule:
            types = rule.split(',')
        else:
            types = rule.split(' AND ')
        rule_entry: list[Rule] = []
        for type_ in types:
            spec_ = type_.split(':')
            if len(spec_) == 1:
                if not current_exclusion_type:
                    exclusion_type = default_exclusion_type
                else:
                    exclusion_type = str(current_exclusion_type)
                expression_position = 0
            else:
                exclusion_type = spec_[0]
                expression_position = 1
            current_exclusion_type = str(exclusion_type)
            exclusion_rule = spec_[expression_position].rstrip()
            rule_entry.append(
                Rule(exclusion_type.strip(), exclusion_rule.strip()))
        rules.append(rule_entry)
    return rules


def _format_raw_rules(raw_rules: str) -> list[str]:
    """Applies formatting to raw string rule."""
    # Removes brackets.
    raw_rules = re.sub('[(|)]', '', raw_rules)
    # Adds whitespace in front of operators.
    raw_rules = re.sub('([>|<|>=|<=|!=|=])', r' \1 ', raw_rules)
    # Trims extra whitespace.
    raw_rules = re.sub(' +', ' ', raw_rules)
    # Split rules by OR operator.
    return raw_rules.split(' OR ')



from __future__ import annotations

import pathlib
from itertools import chain

from setuptools import find_packages
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

EXTRAS_REQUIRE = {
    'gcp': [
        'google-cloud-firestore', 'google-cloud-pubsub',
        'appengine-python-standard>=1.0.0'
    ]
}
EXTRAS_REQUIRE['full'] = list(set(chain(*EXTRAS_REQUIRE.values())))

setup(
    name='googleads-housekeeper',
    version='0.1.0.dev18',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Google Inc. (gTech gPS CSE team)',
    author_email='no-reply@google.com',
    license='Apache 2.0',
    python_requires='>3.8',
    packages=find_packages(),
    install_requires=[
        'google-ads-api-report-fetcher>=1.12.0', 'SQLAlchemy==1.4.46',
        'beautifulsoup4', 'google-api-python-client', 'psycopg2-binary',
        'croniter', 'aiohttp', 'slack_sdk'
    ],
    extras_require=EXTRAS_REQUIRE,
    setup_requires=['pytest-runner'],
    tests_requires=['pytest', 'pytest-mock'],
)



from __future__ import annotations

import gaarf
import pytest

from googleads_housekeeper.adapters import notifications


class TestNotifications:

    @pytest.fixture
    def empty_payload(self):
        return notifications.MessagePayload(
            task_id='450c7bc6-5fe1-402c-9c23-53f1818903e1',
            task_name='test_task',
            task_formula='GOOGLE_ADS_INFO:impressions > 10',
            placements_excluded_sample=None,
            total_placement_excluded=0,
            recipient='test_recipient')

    @pytest.mark.parametrize(
        'notification_type, notification_class',
        [('slack', notifications.SlackNotifications),
         ('email', notifications.GoogleCloudAppEngineEmailNotifications),
         ('console', notifications.ConsoleNotifications),
         ('unknown', notifications.NullNotifications)])
    def test_create_notification_service_creates_correct_notification_class(
            self, notification_type, notification_class):
        notification_service = notifications.NotificationFactory(
        ).create_nofication_service(notification_type=notification_type)
        assert isinstance(notification_service, notification_class)

    def test_sendnotification_raises_error_for_unknown_notification_type(
            self, empty_payload):
        notification_service = notifications.NullNotifications(
            notification_type='unknown')
        with pytest.raises(ValueError):
            notification_service.send(empty_payload)


class TestGoogleCloudAppEngineEmailNotifications:

    def test_prepare_message_returns_correct_message_elements_for_empty_payload(
            self):
        empty_payload = notifications.MessagePayload(
            task_id='450c7bc6-5fe1-402c-9c23-53f1818903e1',
            task_name='test_task',
            task_formula='GOOGLE_ADS_INFO:impressions > 10',
            placements_excluded_sample=None,
            total_placement_excluded=0,
            recipient='test_recipient')
        sender = notifications.GoogleCloudAppEngineEmailNotifications(
            project_id='fake-project')
        message = sender._prepare_message(empty_payload)
        assert message.body == ''
        assert message.to == ['test_recipient']
        assert message.subject == 'No placements were excluded'
        assert message.sender == (
            'Placements_Excluder___test_task@fake-project.appspotmail.com')

    def test_prepare_message_returns_correct_message_elements_for_non_empty_payload(
            self):
        excluded_placements = gaarf.report.GaarfReport(
            results=[[
                'test-placement',
                'example.com',
                'GOOGLE_ADS_INFO:impressions > 10',
            ]],
            column_names=['name', 'placement', 'reason'])
        payload_with_placements = notifications.MessagePayload(
            task_id='450c7bc6-5fe1-402c-9c23-53f1818903e1',
            task_name='test_task',
            task_formula='GOOGLE_ADS_INFO:impressions > 10',
            placements_excluded_sample=excluded_placements,
            total_placement_excluded=1,
            recipient='test_recipient')
        sender = notifications.GoogleCloudAppEngineEmailNotifications(
            project_id='fake-project')
        message = sender._prepare_message(payload_with_placements)
        assert message.to == ['test_recipient']
        assert message.subject == '1 placements were excluded'
        assert message.sender == (
            'Placements_Excluder___test_task@fake-project.appspotmail.com')
        assert 'test-placement' in message.html
        assert 'example.com' in message.html



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for defining execution objects."""
from __future__ import annotations

import dataclasses
import datetime
import enum
import uuid


class ExecutionTypeEnum(enum.Enum):
    """Holds type of Task execution."""
    SCHEDULED = 0
    MANUAL = 1


@dataclasses.dataclass
class Execution:
    """Holds information of a particular task run.

    Attributes:
        task: Id of a task.
        start_date: Time when execution started.
        end_date: Time when execution ended.
        placements_excluded: Number of excluded placements.
        type: Type of execution (MANUAL, SCHEDULED).
        id: Unique identifier of execution.
    """
    task: int
    start_time: datetime.datetime
    end_time: datetime.datetime
    placements_excluded: int
    type: ExecutionTypeEnum
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self) -> None:
        """Ensures that strings are always cast to proper enum."""
        self.type = self._cast_to_enum(ExecutionTypeEnum, self.type)

    def _cast_to_enum(self, enum_: type[enum.Enum],
                      value: str | enum.Enum) -> enum.Enum:
        """Cast to proper enum."""
        return enum_[value] if isinstance(value, str) else value


@dataclasses.dataclass
class ExecutionDetails:
    """Holds detailed information on a particular execution.

    Attributes:
        execution_id: Unique identifier of an execution.
        placement: Placement identifier that was excluded during the execution.
        placement_type: Type of excluded placement.
        reason: Reason for exclusion.
        id: Unique identifier of execution details entry.
    """
    execution_id: str
    placement: str
    placement_type: str
    reason: str
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for specifying base classes for concrete queries.

Ads Housekeeper can perform changes to Google Ads in various ways (i.e. by
excluding placements from accounts). In order to enforce certain rules when
extending the solution various base classes are added so child classes always
contain required fields.
"""
from __future__ import annotations

from gaarf.base_query import BaseQuery


class ExcludableEntity(BaseQuery):
    """Specifies fields that form an entity that can be excluded.

    Attributes:
        base_query_text:
            A Gaarf query text template that contains aliases specified
            in `required_fields`.
    Raises:
        ValueError:
            If subclass base_query_text does not contain all required fields.
    """
    base_query_text = ''

    def __init_subclass__(cls):
        required_fields = (
            'placement_type',
            'placement',
        )
        super().__init_subclass__()
        missing_fields: str = []
        missing_fields = [
            field for field in required_fields
            if field not in cls.base_query_text
        ]
        if missing_fields:
            raise ValueError('query_text does not contain required fields: '
                             f'{missing_fields}')



from __future__ import annotations

import pytest
import requests

from googleads_housekeeper.domain.external_parsers import website_parser
from googleads_housekeeper.domain.external_parsers import youtube_data_parser


class TestWebSiteParser:

    @pytest.mark.parametrize('input_url,formatted_url',
                             [('example.com', 'https://example.com'),
                              ('http://example.com', 'http://example.com'),
                              ('https://example.com', 'https://example.com')])
    def test_convert_url(self, input_url, formatted_url):
        url = website_parser.WebSiteParser()._convert_placement_to_url(
            input_url)
        assert url == formatted_url

    @pytest.mark.parametrize('attribute,expected_value',
                             [('is_processed', True),
                              ('title', 'example title'),
                              ('description', 'example description'),
                              ('keywords', 'example keywords')])
    def test_parse_website_success(self, mocker, attribute, expected_value):
        mocker.patch(
            'googleads_housekeeper.domain.external_parsers.website_parser.'
            'WebSiteParser.parse',
            return_value=[
                website_parser.WebsiteInfo(placement='example.com',
                                           title='example title',
                                           description='example description',
                                           keywords='example keywords',
                                           is_processed=True)
            ])

        parser = website_parser.WebSiteParser()
        website_info = parser.parse(['example.com'])
        assert getattr(website_info[0], attribute) == expected_value
        parser.parse.assert_called_once_with(['example.com'])

    @pytest.mark.parametrize('attribute,expected_value',
                             [('is_processed', False), ('title', ''),
                              ('description', ''), ('keywords', '')])
    def test_parse_website_raise_error_empty_result(self, mocker, attribute,
                                                    expected_value):
        mocker.patch('requests.get',
                     side_effect=requests.exceptions.ConnectionError)
        parser = website_parser.WebSiteParser()
        website_info = parser.parse(['non-example.com'])
        assert getattr(website_info[0], attribute) == expected_value


class TestYouTubeChannelParser:

    @pytest.mark.parametrize('attribute,expected_value',
                             [('title', 'example title'),
                              ('description', 'example description'),
                              ('country', 'US'), ('subscriberCount', 1),
                              ('videoCount', 1),
                              ('topicCategories', ('Film', 'Game')),
                              ('country', 'US'), ('viewCount', 10)])
    def test_parse_youtube_channel_success(self, mocker, attribute,
                                           expected_value):
        mocker.patch(
            'googleads_housekeeper.domain.external_parsers.youtube_data_parser.'
            'ChannelInfoParser.parse',
            return_value=youtube_data_parser.ChannelInfo(
                placement='example_channel_id',
                title='example title',
                description='example description',
                country='US',
                subscriberCount=1,
                videoCount=1,
                viewCount=10,
                topicCategories=('Film', 'Game')))

        parser = youtube_data_parser.ChannelInfoParser()
        channel_info = parser.parse('example_channel_id')
        assert getattr(channel_info, attribute) == expected_value
        parser.parse.assert_called_once_with('example_channel_id')


class TestYouTubeVideoParser:

    @pytest.mark.parametrize('attribute,expected_value',
                             [('title', 'example title'),
                              ('description', 'example description'),
                              ('defaultAudioLanguage', 'EN'),
                              ('defaultLanguage', 'EN'), ('likeCount', 10),
                              ('commentCount', 10), ('favouriteCount', 10),
                              ('topicCategories', ('Film', 'Game')),
                              ('madeForKids', False),
                              ('tags', ('tag1', 'tag2')), ('viewCount', 10)])
    def test_parse_youtube_video_success(self, mocker, attribute,
                                         expected_value):
        mocker.patch(
            'googleads_housekeeper.domain.external_parsers.youtube_data_parser.'
            'VideoInfoParser.parse',
            return_value=youtube_data_parser.VideoInfo(
                placement='example_video_id',
                title='example title',
                description='example description',
                defaultLanguage='EN',
                defaultAudioLanguage='EN',
                viewCount=10,
                commentCount=10,
                likeCount=10,
                favouriteCount=10,
                madeForKids=False,
                tags=('tag1', 'tag2'),
                topicCategories=('Film', 'Game')))

        parser = youtube_data_parser.VideoInfoParser()
        video_info = parser.parse('example_video_id')
        assert getattr(video_info, attribute) == expected_value
        parser.parse.assert_called_once_with('example_video_id')
